/*
  mprsplit.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <math.h>
#include <netinet/in.h>

#include "mpr.h"

// the reference coordinates in the map aren't very precise,
// so add offset for adjustment
typedef struct {
  char name[32];
  int x ,y;
} map_adjust_t;

map_adjust_t map_adjust[] = {
  { "top25/bw100", 90,  75 },
  { "top25/bw25n", 90,  95 },
  { "top25/bw25s", 90, 135 },
};

#define MIN(a,b)  (((a)<(b))?(a):(b))
#define MAX(a,b)  (((a)>(b))?(a):(b))

#define DIR_MODE S_IRWXU

// write files with more bytes decoded than expected
#define WRITE_OVERLENGTH  

typedef struct {
  unsigned long address;
  unsigned long length;
} __attribute__((packed)) mpr_tile_index_t;

typedef struct {
  int id;
  char* ellipsoidName;
  double EquatorialRadius;
  double eccentricitySquared;
} ellipsoid_t;

#define PI         3.14159265
#define FOURTHPI   (PI / 4)
#define DEG2RAD(a) ((a) * PI / 180)
#define RAD2DEG(a) ((a) * 180.0 / PI)

//  id, Ellipsoid name, Equatorial Radius, square of eccentricity	
static ellipsoid_t ellipsoid[] = {
  { -1, "Placeholder",                   0, 0.0         },
  {  1, "Airy",                    6377563, 0.00667054  },
  {  2, "Australian National",     6378160, 0.006694542 },
  {  3, "Bessel 1841",             6377397, 0.006674372 },
  {  4, "Bessel 1841 (Nambia} ",   6377484, 0.006674372 },
  {  5, "Clarke 1866",             6378206, 0.006768658 },
  {  6, "Clarke 1880",             6378249, 0.006803511 },
  {  7, "Everest",                 6377276, 0.006637847 },
  {  8, "Fischer 1960 (Mercury} ", 6378166, 0.006693422 },
  {  9, "Fischer 1968",            6378150, 0.006693422 },
  { 10, "GRS 1967",                6378160, 0.006694605 },
  { 11, "GRS 1980",                6378137, 0.00669438  },
  { 12, "Helmert 1906",            6378200, 0.006693422 },
  { 13, "Hough",                   6378270, 0.00672267  },
  { 14, "International",           6378388, 0.00672267  },
  { 15, "Krassovsky",              6378245, 0.006693422 },
  { 16, "Modified Airy",           6377340, 0.00667054  },
  { 17, "Modified Everest",        6377304, 0.006637847 },
  { 18, "Modified Fischer 1960",   6378155, 0.006693422 },
  { 19, "South American 1969",     6378160, 0.006694542 },
  { 20, "WGS 60",                  6378165, 0.006693422 },
  { 21, "WGS 66",                  6378145, 0.006694542 },
  { 22, "WGS-72",                  6378135, 0.006694318 },
  { 23, "WGS-84",                  6378137, 0.00669438  }
};

void hexdump(void *data, int size) {
  int i,n = 0, b2c;
  char *ptr = data;
  const int wrap = 16;

  if(!size) return;

  while(size>0) {
    printf("%04x: ", n);

    b2c = (size>wrap)?wrap:size;

    for(i=0;i<b2c;i++)
      printf("%02x ", 0xff&ptr[i]);

    printf("  ");

    for(i=0;i<(wrap-b2c);i++)
      printf("   ");

    for(i=0;i<b2c;i++)
      printf("%c", isprint(ptr[i])?ptr[i]:'.');

    printf("\n");

    ptr  += b2c;
    size -= b2c;
    n    += b2c;
  }
}

void file_read(FILE *file, void *addr, int len) {
  if(fread(addr, 1, len, file) != len) {
    perror("read()");
    exit(-1);
  }
}

unsigned long get_bits(unsigned char *data, int startbit, int bits) {
  int byte_index = startbit / 8;
  int bit_index = startbit % 8; 

  // read first byte
  unsigned long result = (0xff >> bit_index) & data[byte_index++];
  bits -= 8-bit_index;

  // read more bytes if required
  while(bits > 0) {
    result = (result<<8) | data[byte_index++];
    bits -= 8;
  }

  if(bits < 0)
    result >>= -bits;

  return result;
}

// segment info from mpr file
typedef struct {
  int color;
  int length;
} segment_t;

// a single "character" within the lzw string
typedef struct {
  unsigned char marker, chr;
} lzw_char_t;

// a lzw string used by the lzw dictionary entry
typedef struct {
  int length;
  lzw_char_t *chars;
} lzw_string_t;

// append a single char to lzw string no i in the string table
void lzw_append_chr(lzw_string_t *string, lzw_char_t chr) { 
  // extent space for chars to hold one more lzw character
  string->chars = realloc(string->chars, 
		   sizeof(lzw_char_t) * (string->length+1));
  // and append new char
  string->chars[string->length] = chr;
  // adjust strings length field
  string->length++;
}

// convert two unsigned chars into lzw_char and
// append it to lzw string no i in the string table
void lzw_append_raw_chr(lzw_string_t *string, 
			unsigned char marker, unsigned char chr) { 
  lzw_char_t lzw_chr = { marker, chr };
  lzw_append_chr(string, lzw_chr);
}

// copy one lzw string into the control structure of a different one
void lzw_copy_string(lzw_string_t *dst, lzw_string_t *src) {
  // alloc space for new string
  dst->chars = realloc(dst->chars, sizeof(lzw_char_t) * src->length);
  memcpy(dst->chars, src->chars, sizeof(lzw_char_t) * src->length);
  dst->length = src->length;
}

int lzw_chr_equal(lzw_char_t a, lzw_char_t b) {
  return((a.marker == b.marker)&&(a.chr == b.chr));
}

// write a single pixel using he color table
void lzw_write_pixel(FILE *file, color_t *colors, int cindex) {
  // this routine only supports 16 color images, thus we write two
  // pixels from one 8 bit color index
  fputc(colors[cindex&15].r, file);
  fputc(colors[cindex&15].g, file);
  fputc(colors[cindex&15].b, file);

  fputc(colors[cindex>>4].r, file);
  fputc(colors[cindex>>4].g, file);
  fputc(colors[cindex>>4].b, file);
}

// write an item (color or segment) into file
void lzw_write_item(FILE *file, color_t *colors, 
		    segment_t *segments, int item) {
  if(item < 256) {  // direct color pixel
    lzw_write_pixel(file, colors, item);
  } else {
    int j;
    for(j=0;j<segments[item-256].length;j++) 
      lzw_write_pixel(file, colors, segments[item-256].color);
  }
}

void lzw_decode(FILE *file,
		unsigned char *code, int nelin, 
		color_t *colors, segment_t *segments, int numseg) {

  int i, j;
  int bitl = 9;
  const int numcol = 256;

  // since numseg is max 256, the whole table has max 512 entries
  int numtab = numcol + numseg;   // actual base table size

  int next_code = numtab;                       // starting code number
  int old = get_bits(code, 0, bitl);            // first code

  // construct start 2-letter string table
  int tablesize = 1<<bitl;

  //IDL:  table = STRARR(tablesize);
  // max 12 bits -> max 4096 table entries
  lzw_string_t table[4096];
  bzero(table, sizeof(table));  // erase it

  // base table strings start with marker 0
  for(i=1;i<numcol;i++) lzw_append_raw_chr(&table[i], 0, i);

  // addtional table strings start with marker 1
  for(i=numcol+1;i<numtab;i++) lzw_append_raw_chr(&table[i], 1, i-numcol);

  // empty strings have to get separate table values:
  lzw_append_raw_chr(&table[0], 2, '0');
  lzw_append_raw_chr(&table[numcol], 3, '0');

  // workaround problem with bw25n/16016 (7303) -> skip them
  if(!table[old].chars) {
    printf("DECODE init error!\n");
    return;
  }

  lzw_char_t chr = table[old].chars[0];

  // initialize output array:
  lzw_write_item(file, colors, segments, old); // first code of output

  lzw_string_t str = { 0, NULL };
  unsigned long startbit = bitl;            // bit pointer into binarr

  // read all bits
  while((startbit + bitl) < nelin) {    
    int new = get_bits(code,startbit,bitl); // extract next code

    if(new >= next_code) {   // check whether code is not in table
      if(old >= next_code) { 
	//IDL: str = str + char $
	lzw_append_chr(&str, chr);
      } else { 
	//IDL: str = table[old]+char
	lzw_copy_string(&str, &table[old]);
	lzw_append_chr(&str, chr);
      }
    } else {
      //IDL: str=table[new]
      lzw_copy_string(&str, &table[new]);
    }
    
    // decompose 'str' for output
    for(i=0;i<str.length;i++) {
      int index = -1;
      
      // search for encoded byte in array 
      //IDL:    index = WHERE(table0 EQ STRMID(str,i,2))
      for(j=0;j<numtab;j++) {
	if(lzw_chr_equal(table[j].chars[0], str.chars[i]))
	  index = j;
      }

      //IDL:   outarr[outc]=index[0]
      //IDL:   outc=outc+1
      lzw_write_item(file, colors, segments, index);
    }
    
    //IDL: char = STRMID(str,0,2)   
    chr = str.chars[0];           // extract start char from str

    if(next_code < tablesize) {          // extend table by new code
      if(old >= next_code) {
        //IDL:      table[next_code]=str+char $
	lzw_copy_string(&table[next_code], &str);
	lzw_append_chr(&table[next_code], chr);
      } else {
	// copy old code to next_code and attach chr
	//IDL: table[next_code]=table[old]+char
	lzw_copy_string(&table[next_code], &table[old]);
	lzw_append_chr(&table[next_code], chr);
      }
      next_code = next_code + 1;
    }
    
    old = new;
    startbit = startbit + bitl;
    
    // if next_code reaches tablesize extend table
    // and switch to one bit longer reading until bitl=12
    if((next_code == tablesize) && (bitl < 12)) {
      // longer codes always start on byte boundary:
      if(bitl < 12) startbit = ((startbit+7)/8)*8;

      bitl++;
      tablesize = 1<<bitl;
    }
  }
  
  // free all strings
  for(i=0;i<4096;i++) 
    if(table[i].chars)
      free(table[i].chars);
}

char UTMLetterDesignator(double Lat) {
  // This routine determines the correct UTM letter designator for the 
  // given latitude returns 'Z' if latitude is outside the UTM limits 
  // of 84N to 80S
  // Written by Chuck Gantz- chuck.gantz@globalstar.com
  char LetterDesignator;

  if((84 >= Lat) && (Lat >= 72)) LetterDesignator = 'X';
  else if((72 > Lat) && (Lat >= 64)) LetterDesignator = 'W';
  else if((64 > Lat) && (Lat >= 56)) LetterDesignator = 'V';
  else if((56 > Lat) && (Lat >= 48)) LetterDesignator = 'U';
  else if((48 > Lat) && (Lat >= 40)) LetterDesignator = 'T';
  else if((40 > Lat) && (Lat >= 32)) LetterDesignator = 'S';
  else if((32 > Lat) && (Lat >= 24)) LetterDesignator = 'R';
  else if((24 > Lat) && (Lat >= 16)) LetterDesignator = 'Q';
  else if((16 > Lat) && (Lat >= 8)) LetterDesignator = 'P';
  else if(( 8 > Lat) && (Lat >= 0)) LetterDesignator = 'N';
  else if(( 0 > Lat) && (Lat >= -8)) LetterDesignator = 'M';
  else if((-8> Lat) && (Lat >= -16)) LetterDesignator = 'L';
  else if((-16 > Lat) && (Lat >= -24)) LetterDesignator = 'K';
  else if((-24 > Lat) && (Lat >= -32)) LetterDesignator = 'J';
  else if((-32 > Lat) && (Lat >= -40)) LetterDesignator = 'H';
  else if((-40 > Lat) && (Lat >= -48)) LetterDesignator = 'G';
  else if((-48 > Lat) && (Lat >= -56)) LetterDesignator = 'F';
  else if((-56 > Lat) && (Lat >= -64)) LetterDesignator = 'E';
  else if((-64 > Lat) && (Lat >= -72)) LetterDesignator = 'D';
  else if((-72 > Lat) && (Lat >= -80)) LetterDesignator = 'C';

  // This is here as an error flag to show that the Latitude is 
  // outside the UTM limits
  else LetterDesignator = 'Z';

  return LetterDesignator;
}

void LLtoUTM(int ReferenceEllipsoid, const double Lat, const double Long,
             long *UTMNorthing, long *UTMEasting, unsigned short *UTMZone, 
	     char* UTMBand) {

  // converts lat/long to UTM coords.  Equations from USGS Bulletin 1532
  // East Longitudes are positive, West longitudes are negative.
  // North latitudes are positive, South latitudes are negative
  // Lat and Long are in decimal degrees
  // Written by Chuck Gantz- chuck.gantz@globalstar.com

  double a = ellipsoid[ReferenceEllipsoid].EquatorialRadius;
  double eccSquared = ellipsoid[ReferenceEllipsoid].eccentricitySquared;
  double k0 = 0.9996;

  double LongOrigin;
  double eccPrimeSquared;
  double N, T, C, A, M;

  // Make sure the longitude is between -180.00 .. 179.9
  double LongTemp = (Long+180)- (int)((Long+180)/360)*360-180;

  double LatRad = DEG2RAD(Lat);
  double LongRad = DEG2RAD(LongTemp);
  double LongOriginRad;

  *UTMZone = (int)((LongTemp + 180)/6) + 1;

  if( Lat >= 56.0 && Lat < 64.0 && LongTemp >= 3.0 && LongTemp < 12.0 )
    *UTMZone = 32;

  // Special zones for Svalbard
  if( Lat >= 72.0 && Lat < 84.0 ) {
    if(      LongTemp >= 0.0  && LongTemp <  9.0 ) *UTMZone = 31;
    else if( LongTemp >= 9.0  && LongTemp < 21.0 ) *UTMZone = 33;
    else if( LongTemp >= 21.0 && LongTemp < 33.0 ) *UTMZone = 35;
    else if( LongTemp >= 33.0 && LongTemp < 42.0 ) *UTMZone = 37;
  }

  //+3 puts origin in middle of zone
  LongOrigin = (*UTMZone - 1)*6 - 180 + 3;
  LongOriginRad = DEG2RAD(LongOrigin);

  *UTMBand = UTMLetterDesignator(Lat);

  eccPrimeSquared = (eccSquared)/(1-eccSquared);

  N = a/sqrt(1-eccSquared*sin(LatRad)*sin(LatRad));
  T = tan(LatRad)*tan(LatRad);
  C = eccPrimeSquared*cos(LatRad)*cos(LatRad);
  A = cos(LatRad)*(LongRad-LongOriginRad);

  M = a*((1 - eccSquared/4 - 3*eccSquared*eccSquared/64 -
          5*eccSquared*eccSquared*eccSquared/256)*LatRad
         - (3*eccSquared/8 + 3*eccSquared*eccSquared/32 +
            45*eccSquared*eccSquared*eccSquared/1024)*sin(2*LatRad)
         + (15*eccSquared*eccSquared/256 +
            45*eccSquared*eccSquared*eccSquared/1024)*sin(4*LatRad)
         - (35*eccSquared*eccSquared*eccSquared/3072)*sin(6*LatRad));

  *UTMEasting = k0*N*(A+(1-T+C)*A*A*A/6
               + (5-18*T+T*T+72*C-58*eccPrimeSquared)*A*A*A*A*A/120)
               + 500000.0;

  *UTMNorthing = k0*(M+N*tan(LatRad)*(A*A/2+(5-T+9*C+4*C*C)*A*A*A*A/24
                 + (61-58*T+T*T+600*C-330*eccPrimeSquared)*A*A*A*A*A*A/720));

  // 10000000 meter offset for southern hemisphere
  if(Lat < 0)
    *UTMNorthing += 10000000.0;
}

int mpr_load_header(char *name, map_info_t *map_info) {
  char map_name[32];
  char file_name[128];
  FILE *file;
  int i;
  unsigned short dummy16;

  strcpy(file_name, name);
  strcat(file_name, ".mph");

  printf("\n>>>>>>>>>>>>>>>>> Parsing header file %s <<<<<<<<<<<<<<<<<<<\n", file_name);
  
  file = fopen(file_name, "rb");
  if(!file) {
    perror("opening mph file");
    return -1;
  }
  
  file_read(file, &map_info->mph_header, sizeof(mph_header_t));

  //  printf("marker = %ld\n", map_info->mph_header.marker);
  printf("tile size = %ld*%ld\n", 
	 map_info->mph_header.tile_width, map_info->mph_header.tile_height);
  printf("tiles = %ld*%ld (%lu)\n", 
	 map_info->mph_header.tiles_x, map_info->mph_header.tiles_y,
	 map_info->mph_header.tiles_x * map_info->mph_header.tiles_y);

  if(map_info->mph_header.tile_width != map_info->mph_header.tile_height) {
    printf("ERROR: tile size height and width differ\n");
    return -1;
  }
    
  if((map_info->mph_header.tile_width != 256) &&
     (map_info->mph_header.tile_width != 128)) {
    printf("ERROR: tile size not 256x256 or 128x128\n");
    return -1;
  }

  map_info->width  = map_info->mph_header.tiles_x;
  map_info->height = map_info->mph_header.tiles_y;

  // map name is stored at 596th byte
  fseek(file, 596, SEEK_SET);
  file_read(file, map_name, sizeof(map_name));
  printf("Map name: \"%s\"\n", map_name);

  strncpy(map_info->name, map_name, sizeof(map_info->name));

  // seek to even address behind name
  fseek(file, 596+strlen(map_name)+1+((strlen(map_name)+1)&1), SEEK_SET);

  // try to find valid pair of scale values
  // search for 16 bit value 130 followed by 16 bit value 4
  dummy16 = 0;
  while(dummy16 != 130) {
    file_read(file, &dummy16, sizeof(dummy16));

    // short 130 must be immediately followed by 4
    if(dummy16 == 130) {
      dummy16 = 0;
      file_read(file, &dummy16, sizeof(dummy16));
      if(dummy16 == 4) dummy16 = 130;
    }
  }

  //  printf("130/4 scan ok\n");

  unsigned long scale[2];
  file_read(file, scale, sizeof(scale));

  printf("scale: %lu:%lu\n", scale[0], scale[1]);

  map_info->scale[0] = scale[0];
  map_info->scale[1] = scale[1];

  // skip 2 unknown longs
  fseek(file, 2*sizeof(long), SEEK_CUR);

  double dpmm;  // dots per millimeter
  file_read(file, &dpmm, sizeof(dpmm));
  printf("dots/mm: %.2f\n", dpmm);

  unsigned long cal_points[2]; // number of calibration points
  file_read(file, cal_points, sizeof(cal_points));
  printf("cal_points: (%lu) %lu\n", cal_points[0], cal_points[1]);
  // meaning of cal_points[0] unknown

  // read all calibration points
  for(i=0;i<cal_points[1];i++) {
    printf("Calibration point %u:\n", i);

    unsigned long pixel[4];  // pixel pairs
    file_read(file, pixel, sizeof(pixel));
    printf("  pixel pairs: %lu %lu %lu %lu\n",
	   pixel[0], pixel[1], pixel[2], pixel[3]);

    double coord[2];  // coordinate pair
    file_read(file, coord, sizeof(coord));
    printf("  coord: %f %f\n", coord[0], coord[1]);

    // process first pair only
    if(i == 0) {    
      int adj;
      unsigned short dmperpixel = scale[1] / 100 / dpmm;
      long northing, easting;

      // easting and nothing are in meters
      LLtoUTM(23, coord[0], coord[1], &northing, &easting, 
	      &map_info->zone, &map_info->band);
      printf("  utm: %d%c E %ld N %ld\n", map_info->zone, map_info->band, 
	     easting, northing); 

      // adjust pixel coordinates to meters
      printf("  dm per pixel: %u\n", dmperpixel);
      map_info->dm_per_pixel = dmperpixel;

      pixel[0] = pixel[0] * dmperpixel / 10;
      pixel[1] = (map_info->mph_header.tile_height * map_info->mph_header.tiles_y - pixel[1]) 
	* dmperpixel / 10;

      printf("  pixel to meters: %ld %ld\n", pixel[0], pixel[1]);

      // calculate offset between map and utm in meters
      pixel[0] = easting - pixel[0];
      pixel[1] = northing - pixel[1];

      printf("  map offset in meters: %ld %ld\n", pixel[0], pixel[1]);

      // check for offset adjustment
      for(adj=0;adj<sizeof(map_adjust)/sizeof(map_adjust_t);adj++) {
	if(strcasecmp(map_adjust[adj].name, name) == 0) {
	  printf("  map adjustment found: %d and %d meters\n", 
		 map_adjust[adj].x, map_adjust[adj].y);

	  pixel[0] -= map_adjust[adj].x;
	  pixel[1] -= map_adjust[adj].y;
	}
      }

      map_info->x_offset = pixel[0];
      map_info->y_offset = pixel[1];
    }

    // skip 32 unknown bytes
    fseek(file, 32l, SEEK_CUR);
  }
  
  // search for 16 bit value 150 followed by 16 bit value 3
  dummy16 = 0;
  while(dummy16 != 150) {
    file_read(file, &dummy16, sizeof(dummy16));

    // short 150 must be immediately followed by 4
    if(dummy16 == 150) {
      dummy16 = 0;
      file_read(file, &dummy16, sizeof(dummy16));
      if(dummy16 == 3) dummy16 = 150;
    }
  }

  //  printf("150/3 scan ok\n");

  // read color table
  unsigned short coltab_head[7];
  file_read(file, coltab_head, sizeof(coltab_head));

  map_info->colors = malloc(sizeof(color_t) * coltab_head[2]);

  for(i=0;i<coltab_head[2];i++) {
    unsigned short rgb[3];
    file_read(file, rgb, sizeof(rgb));

    // store colors in color table
    map_info->colors[i].r = rgb[0];
    map_info->colors[i].g = rgb[1];
    map_info->colors[i].b = rgb[2];
  }

  // mph file completely parsed
  fclose(file);

  printf("MPH header ok\n");

  return 0;
}

int mpr_get(char *path, char *name, map_info_t *map_info, int tile_no) {
  char file_name[128];
  segment_t *segments = NULL;
  FILE *file;
  int i, verbose=0;

  sprintf(file_name, "%s.mpr", map_info->filename);

  if(verbose)
    printf(">>>>>> Reading map file %s <<<<<<\n", file_name);
  
  file = fopen(file_name, "rb");
  if(!file) {
    perror("opening mpr file");
    return -1;
  }

  // the file begins with a table of 32 bit address/length pairs  
  int tiles = map_info->mph_header.tiles_x * map_info->mph_header.tiles_y;
  mpr_tile_index_t *mpr_tile_index = malloc(sizeof(mpr_tile_index_t) * tiles);

  if(verbose)
    printf("Reading index for %lu*%lu=%u tiles\n", 
	   map_info->mph_header.tiles_x, map_info->mph_header.tiles_y, tiles);

  // read address/length pairs for each tile
  file_read(file, mpr_tile_index, sizeof(mpr_tile_index_t) * tiles);

  if(verbose)
    printf("Request to read tile no %d\n", tile_no);

  if(tile_no > tiles) {
    printf("tile number too big\n");
    return -1;
  }

  if(!mpr_tile_index[tile_no].address) {
    printf("Tile %d is empty\n", tile_no);
  } else {
    
    // read whole tile
    fseek(file, mpr_tile_index[tile_no].address, SEEK_SET);
    unsigned char *tile = malloc(mpr_tile_index[tile_no].length);
    file_read(file, tile, mpr_tile_index[tile_no].length);
    
    // segments in first byte
    segments = malloc(sizeof(segment_t) * tile[0]);
    
    int startbit = 24;
    int lastcol, thiscol, thislen;
    int bit2 = 2;
    int bitmax = 8;
    
    segments[0].color = lastcol = thiscol = get_bits(tile, 8, 8);
    segments[0].length = thislen = get_bits(tile, 16, 8);
    
    for(i=1;i<tile[0];i++) {
      thiscol = get_bits(tile,startbit,8);  // next color value, always 8 bit
      startbit = startbit+8;
      
      // read next length value only if previous was larger than 2
      if(thislen > 2) {
	// if first bit of next length value is 1, stick with previous 
	// lenghth, except color is the same
	if((thislen <= 128) && (get_bits(tile, startbit, 1) == 1) &&
	   (thiscol != lastcol)) 
	  startbit++;
	else {	
	  // read next length value bit by bit until larger than previous one
	  // then go back 1 bit
	  while((get_bits(tile,startbit,bit2+1) <= thislen) &&
		(bit2 < bitmax)) 
	    bit2 = bit2 + 1;
	  
	  // if length and color are same as before, decrease length
	  if((get_bits(tile,startbit,bit2) == thislen) && 
	     (thiscol == lastcol))
	    bit2 = bit2-1;
	  
	  thislen = MAX(get_bits(tile,startbit,bit2),2);
	  
	  // if length=2 only color values follow, but start at full byte 
	  // boundary
	  if(thislen == 2) startbit = ((startbit + bit2 - 1)/8)*8+8;
	  else             startbit = startbit + bit2;
	  
	  bitmax = MIN(bit2,8);
	  if(bit2 > 2) 
	    while((1<<bitmax) > thislen)
	      bitmax = bitmax-1;
	  bitmax = MIN(bitmax+2,8);
	  bit2 = 2;
	}
      }
      lastcol = thiscol;
      
      segments[i].color = thiscol;
      segments[i].length = thislen;
    }
    
    // get address of first lzw byte
    unsigned char *lzw_data = tile + (startbit+7)/8;

    // open raw output file
    char filename[128];
    sprintf(filename, "%s/%s/%03ld/%03ld.raw", path, name,
	    tile_no / map_info->mph_header.tiles_x,
	    tile_no % map_info->mph_header.tiles_x);

    if(verbose)
      printf("Writing tile %d/%ld to %s\n", tile_no, 
	     map_info->mph_header.tiles_x * map_info->mph_header.tiles_y, 
	     filename);

    FILE *out = fopen(filename, "wb");
    
    lzw_decode(out, 
	       // pointer to source data and number of lzw data bits
	       lzw_data, (mpr_tile_index[tile_no].length - (startbit+7)/8)*8, 
	       // pointer to colors and segment table and its length
	       map_info->colors, segments, tile[0]);
    
    //      printf("written: %d bytes\n", ftell(out));
    
    // file may need to be filled up with white pixels
    while(ftell(out) < 3 * map_info->mph_header.tile_width * 
	  map_info->mph_header.tile_height)
      fputc(0xff, out);
    
    if(ftell(out) > 3 * map_info->mph_header.tile_width * 
       map_info->mph_header.tile_height) {
      printf("ERROR: too many bytes written: %ld > %ld\n", 
	     ftell(out), 3 * map_info->mph_header.tile_width * 
	     map_info->mph_header.tile_height);
      
      // the decoded image is wrong and so is the same image with the
      // original IDL script -> just delete the file ...
      
      fclose(out);
    } else {
      fclose(out);

      // call external convert tool
      char cmd[128];
      sprintf(cmd, "convert -depth 8 -size %ldx%ld "
	      "RGB:%s %s/%s/%03ld/%03ld.gif", 
	      map_info->mph_header.tile_width, 
	      map_info->mph_header.tile_height, 
	      filename, path, name,
	      tile_no / map_info->mph_header.tiles_x,
	      tile_no % map_info->mph_header.tiles_x);

      if(verbose) printf("Calling %s\n", cmd);
      
      if(system(cmd) == -1) {
	perror("calling convert");
	return -1;
      }
    }
    
    // remoce raw file
    if(verbose) printf("removing %s\n", filename);
    remove(filename);
  }

  return 0;
}

int mpr_utm2map(map_info_t *map_info, 
		int northing, int easting, short zone, char band, 
		int *x, int *y) {

  if(zone != map_info->zone) {
    short ref_zone;
    char ref_band;
    //    double latitude, longitude;

    // build reference zone
    ref_zone = map_info->zone;
    ref_band = map_info->band;

    printf("MAP: problem: UTM zone mismatch (%d%c != map %d%c), "
	   "forcing zone to map\n", 
	   zone, band, map_info->zone, map_info->band);

#if 0    
    // try to force into current utm zone by converting
    // to lat/long and then back to utm with forced zone
    utm_utm2ll(utm, &latitude, &longitude);
    utm_ll2utm(latitude, longitude, utm, &ref);
#endif
  }

#if 0
  {
    char band_str[2];
    band_str[0] = utm->band;
    band_str[1] = 0;
    DEBUG_PRINTF("MAP: utm = %d%s E %ld N %ld\n", 
		 utm->zone, band_str, utm->easting, utm->northing);  
  }
#endif

  printf("MAP: %d dm/pixel\n", map_info->dm_per_pixel);

  // convert to UTM offsets
  *x = easting - map_info->x_offset; 
  *y = northing - map_info->y_offset;

  printf("MAP: map meters = %d/%d\n", *x, *y);

  // convert to meters and adjust vertical orientation
  *x = (*x * 10) / map_info->dm_per_pixel;
  *y = (256l * map_info->height) - ((*y * 10) / map_info->dm_per_pixel);

  printf("MAP: map pixel = %d/%d\n", *x, *y);

  // check if this is within the map
  if((*x < 0) || (*x >= 256l * map_info->width+1) ||
     (*y < 0) || (*y >= 256l * map_info->height+1)) {
    printf("MAP: utm coordinate outside map\n");
    return -1;
  }

  return 0;
}

int mpr_map2utm(map_info_t *map_info, 
		int x, int y,
		int *northing, int *easting, short *zone, char *band) {
#if 1
  // check if this is within the map
  if((x < 0) || (x >= 256l * map_info->width+1) ||
     (y < 0) || (y >= 256l * map_info->height+1)) {
    printf("Haeh? utm coordinate outside map?\n");
    return -1;
  }
#endif
  // convert to meters and adjust vertical orientation
  x = (x * map_info->dm_per_pixel)/10;
  y = (((256l * map_info->height)-y)*map_info->dm_per_pixel)/10;

  // convert to UTM offsets
  *easting = x + map_info->x_offset;
  *northing = y + map_info->y_offset;

  if(zone) *zone = map_info->zone;
  if(band) *band = map_info->band;

  return 0;
}
