/*
 * Copyright (C) 2008 Till Harbaum <till@harbaum.org>
 *
 * This file is part of topsrv.
 *
 * Toposrv is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Maemo Mapper is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Maemo Mapper.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Cache layout:
 * cachedir/mapname/y/x.gif   ->  original tiles from map mapname
 * cachedir/tmp*.gif          ->  temporary files used during conversion
 * cachedir/zoom/y/x.gif      ->  rendered tiles in maemo-mapper format
 *
 * http://172.16.61.99:8181/13/4290/2814.gif
 */


#define EXT  "gif"
#define SCALE 12.5  // 12.625

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <math.h>
#include "mpr.h"

#define BUFSIZE 8192

#include <sys/stat.h>

/* make sure the entire path "dir" exists and create it if not */
int checkdir(char *dir) {
  struct stat filestat;
  char *p = dir, tmp;

  do {
    while(*p && *p != '/') p++;
    tmp = *p;
    *p = 0;

    int err = stat(dir, &filestat);
    if(err) {
      if(mkdir(dir, S_IRWXU) != 0) {
	perror("mkdir()");
	return -1;
      }
    } else {
      if(!filestat.st_mode & S_IFDIR) {
	printf("File %s exists and is _no_ directory\n", dir);
	return -1;
      }
    }
    
    *p++ = tmp;
  } while(tmp);

  return 0;
}

void send404(int fd, char *msg) {
  char buffer[256];

  sprintf(buffer,
	  "HTTP/1.0 404 Not found\r\n"
	  "Content-Type: text/html\r\n\r\n"
	  "<html><body><h1>404 - Not found</h1>%s</body></html>", msg);
  write(fd,buffer,strlen(buffer));
  close(fd);
}

/* this web server function serves a single http request */
/* it is called whenever a new connection is being established to */
/* the server */
int web(map_info_t *map_info, int fd) {
  long i, ret;
  static char buffer[BUFSIZE+1]; /* static so zero filled */

  ret = read(fd,buffer,BUFSIZE);   /* read Web request in one go */
  if(ret == 0 || ret == -1) {     /* read failure stop now */
    printf("failed to read browser request\n");
    perror("read()");
  }

  if(ret > 0 && ret < BUFSIZE)    /* return code is valid chars */
    buffer[ret]=0;          /* terminate the buffer */
  else 
    buffer[0]=0;

  for(i=0;i<ret;i++)      /* terminate at CF and LF characters */
    if(buffer[i] == '\r' || buffer[i] == '\n')
      buffer[i]='*';
  
  if( strncmp(buffer,"GET ",4) && strncmp(buffer,"get ",4) ) {
    printf("Only GET operation supported\n");
    send404(fd, "Only GET operation supported");
    return -1;
  }

  /* seperate requested URL */
  for(i=4;i<BUFSIZE;i++) {
    if(buffer[i] == ' ') {
      buffer[i] = 0;
      break;
    }
  }
  
  // analyze URL to match openstreetmap request
  int zoom, x,y;
  i = sscanf(buffer+4, "/%d/%d/%d." EXT, &zoom, &x, &y);
  if(i!=3) {
    printf("Unsupported URL requested: %s\n", buffer+4);
    send404(fd, "Unsupported URL format");
    return -1;
  }

  printf("Request for zoom=%d, x=%d, y=%d\n", zoom, x, y);

  // display real life coordinates for this
  double n = M_PI - 2*M_PI * y/pow(2,zoom);
  double latitude = (180.0/M_PI*atan(0.5*(exp(n)-exp(-n))));
  double longitude = (x/pow(2,zoom)*360.0)-180.0;

  // easting and northing are in meters
  long northing, easting;
  unsigned short zone;
  char band;
  LLtoUTM(WGS84, latitude, longitude, &northing, &easting, &zone, &band);
  printf("  utm: %d%c E %ld N %ld\n", zone, band, easting, northing); 
 
  /* search through all maps ... */
  while(map_info->name[0] && 
	((zoom != map_info->zoom) ||
	 (band != map_info->band) || (zone != map_info->zone) ||
	 (easting < map_info->min_easting) || (easting > map_info->max_easting) ||
	 (northing < map_info->min_northing || northing > map_info->max_northing))) {

#if 1 // show failure details
    printf("in map %s\n", map_info->name);
    printf("Bad zoom or bad position:\n");
    printf("Map: Zoom %d %d%c E %d-%d N %d-%d\n", 
	   map_info->zoom, map_info->zone, map_info->band, 
	   map_info->min_easting, map_info->max_easting,
	   map_info->min_northing, map_info->max_northing); 

    if(zoom != map_info->zoom) printf("zoom failed\n");
    if(band != map_info->band || zone != map_info->zone)  printf("band/zone failed\n");
    if((easting < map_info->min_easting) || (easting > map_info->max_easting)) 
      printf("easting failed\n");
    if((northing < map_info->min_northing) || (northing > map_info->max_northing)) 
      printf("northing failed\n");
#endif

    map_info++;
  }

  if(!map_info->name[0]) {
    printf("No matching map\n");
    send404(fd, "No matching map");
    return -1;
  }

  int file_fd;
  char tmpname[256], filename[256];

  /* check if dir exists */
  sprintf(filename, "%s/%0d/%d", map_info->cachedir, zoom, x);
  if(checkdir(filename) != 0) {
    printf("Unable to create cache dir %s\n", filename);
    send404(fd, "Unable to create cache dir");
    return -1;
  }

  sprintf(filename, "%s/%0d/%d/%d." EXT, map_info->cachedir, zoom, x, y);
  printf("Preparing to send %s\n", filename);

  /* check if file exists ... */
  if(( file_fd = open(filename,O_RDONLY)) == -1) {
    printf("No existing file %s, rendering one ...\n", filename);

    int cx, cy;
    if(mpr_utm2map(map_info, northing, easting, zone, band, &cx, &cy)<0) {
      printf("Position not on map\n");
      /* todo: report error */
      close(fd);
      return -1;
    }
    
    printf(" pixel on map %d %d -> tile %d %d\n", cx, cy, cx/256, cy/256);

    int tx, ty;
    
    /* get all nine adjacent tiles */
    for(ty=0;ty<3;ty++) {
      /* check if dir to place original tile exists */
      sprintf(tmpname, "%s/%s/%03d", map_info->cachedir, 
	      map_info->basename, cy/256+ty);
      if(checkdir(tmpname) != 0) {
	printf("Unable to create cache dir %s\n", filename);
	send404(fd, "Unable to create cache dir");
	return -1;
      }

      for(tx=0;tx<3;tx++) {
	int test_fd;

	/* try to open original tile */
	sprintf(tmpname, "%s/%s/%03d/%03d." EXT, map_info->cachedir, 
		map_info->basename, cy/256+ty, cx/256+tx);
	printf("trying to read %s ...", tmpname);

	if(( test_fd = open(tmpname,O_RDONLY)) == -1) {
	  printf(" failed, regenerating ...");
	  mpr_get(map_info->cachedir, map_info->basename, map_info,
		  (cy/256+ty)*map_info->mph_header.tiles_x+(cx/256)+tx);
	  printf(" ok\n");
	} else {
	  printf(" ok\n");
	  close(test_fd);
	}
      }
    }
    
    /* glue them all nine together (call external convert tool) */
    char cmd[512];
    sprintf(cmd, "montage -geometry +0+0 "
	    "%s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" "
	    "%s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" "
	    "%s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" %s/%s/%03d/%03d."EXT" "
	    "%s/tmp." EXT,
	    map_info->cachedir, map_info->basename, cy/256+0, cx/256+0, 
	    map_info->cachedir, map_info->basename, cy/256+0, cx/256+1, 
	    map_info->cachedir, map_info->basename, cy/256+0, cx/256+2,
	    map_info->cachedir, map_info->basename, cy/256+1, cx/256+0, 
	    map_info->cachedir, map_info->basename, cy/256+1, cx/256+1, 
	    map_info->cachedir, map_info->basename, cy/256+1, cx/256+2,
	    map_info->cachedir, map_info->basename, cy/256+2, cx/256+0, 
	    map_info->cachedir, map_info->basename, cy/256+2, cx/256+1, 
	    map_info->cachedir, map_info->basename, cy/256+2, cx/256+2,
	    map_info->cachedir);
    printf("Calling %s\n", cmd); if(system(cmd) == -1) { 
      perror("calling convert"); close(fd); return -1; }

    /* cut to intra-pic offset */
    sprintf(cmd, "convert %s/tmp."EXT" -crop +%d+%d +repage %s/tmp_crop."EXT, 
	    map_info->cachedir, cx%256, cy%256, map_info->cachedir);
    printf("Calling %s\n", cmd); if(system(cmd) == -1) { 
      perror("calling convert"); close(fd); return -1; }
    
    sprintf(cmd, "convert %s/tmp_crop."EXT" -scale %f%% %s/tmp_scale."EXT, 
	    map_info->cachedir, map_info->rescale, map_info->cachedir);
    printf("Calling %s\n", cmd); if(system(cmd) == -1) { 
      perror("calling convert"); close(fd); return -1; }
    
    sprintf(cmd, "convert %s/tmp_scale."EXT" -crop 256x256+0+0 +repage %s", 
	    map_info->cachedir, filename);
    printf("Calling %s\n", cmd); if(system(cmd) == -1) { 
      perror("calling convert"); close(fd); return -1; }
    
    if(( file_fd = open(filename,O_RDONLY)) == -1) {
      printf("Unable to open file %s\n", filename);
      perror("open()");
      close(fd);
      return -1;
    }
  }

  sprintf(buffer,"HTTP/1.0 200 OK\r\nContent-Type: image/"EXT"\r\n\r\n");
  write(fd,buffer,strlen(buffer));
    
  /* send file in 8KB block - last block may be smaller */
  while ( (ret = read(file_fd, buffer, BUFSIZE)) > 0 ) {
    //    printf("sending %ld bytes\n", ret);
    write(fd,buffer,ret);
  }
  
  close(file_fd);
  close(fd);
  
  return 1;
}

int main(int argc, char **argv) {
  int port, listenfd, socketfd;
  int i;
  size_t length;
  static struct sockaddr_in cli_addr; /* static = initialised to zeros */
  static struct sockaddr_in serv_addr; /* static = initialised to zeros */

  /* structure to hold contents of map.mph file */
  map_info_t *map_info = NULL;

  if(argc < 4) {
    printf("Usage:   toposrv port cachedir map1 [map2...]\n");
    printf("         the mapX.mph and mapX.mpr files will be used\n");
    printf("Example: toposrv 8181 ./cache top25/bw100 top25/bw25n top25/bw25s\n");
    exit(-1);
  }

  /* make space for all map infos (plus an empty last one) */
  map_info = malloc((argc-1)*sizeof(map_info_t));
  memset(map_info, 0, (argc-1)*sizeof(map_info_t));

  /* cut trailing slash from cache dir */
  if(argv[2][strlen(argv[2])-1] == '/')
    argv[2][strlen(argv[2])-1] = 0;

  /* make sure cache dir exists */
  if(checkdir(argv[2]) != 0) {
    printf("Unable to open/create cache dir\n");
    exit(-1);
  }

  /* load all map headers and parse them */
  for(i=3;i<argc;i++) {
    /* load file */
    strcpy(map_info[i-3].filename, argv[i]);
    strcpy(map_info[i-3].cachedir, argv[2]);

    /* get base name */
    char *p = argv[i] + strlen(argv[i]) - 1;
    while(*p != '/' && p > argv[i]) p--;
    if(*p == '/') p++;
    strcpy(map_info[i-3].basename, p);

    if(mpr_load_header(argv[i], &map_info[i-3]) < 0) {
      printf("Unable to load map %s\n", argv[i]);
      exit(-1);
    }

    /* calculate maemo-mapper specific stuff (e.g. zoom) */
    for(map_info[i-3].zoom=0;1024000/(1<<map_info[i-3].zoom)>map_info[i-3].dm_per_pixel;map_info[i-3].zoom++);
    map_info[i-3].zoom--;

    map_info[i-3].rescale = (1<<map_info[i-3].zoom)*map_info[i-3].dm_per_pixel/10240.0;

    printf("Maemo-mapper zoom: %d, rescale by %f%%\n", map_info[i-3].zoom, map_info[i-3].rescale);

    mpr_map2utm(&map_info[i-3], 0, 256l * map_info[i-3].height-1,
		&map_info[i-3].min_northing, &map_info[i-3].min_easting, NULL, NULL);

    mpr_map2utm(&map_info[i-3], 256l * map_info[i-3].width-1, 0,
		&map_info[i-3].max_northing, &map_info[i-3].max_easting, NULL, NULL);

    printf("UTM bounds easting %d %d (%dm), northing %d %d (%dm)\n", 
	   map_info[i-3].min_easting, map_info[i-3].max_easting, map_info[i-3].max_easting-map_info[i-3].min_easting,
	   map_info[i-3].min_northing, map_info[i-3].max_northing, map_info[i-3].max_northing-map_info[i-3].min_northing);
  }

  printf("toposrv starting, pid = %d\n", getpid());

  /* setup the network socket */
  if((listenfd = socket(AF_INET, SOCK_STREAM,0)) <0)
    perror("socket()");

  port = atoi(argv[1]);
  printf("listening at port %d\n", port);

  serv_addr.sin_family = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port = htons(port);
  if(bind(listenfd, (struct sockaddr *)&serv_addr,sizeof(serv_addr)) <0)
    perror("bind()");

  if( listen(listenfd,64) <0)
    perror("listen()");

  for(;;) {
    length = sizeof(cli_addr);
    if((socketfd = accept(listenfd,(struct sockaddr *)&cli_addr, &length)) < 0)
      perror("accept()");

    web(map_info, socketfd);
  }
}
